# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import errno
import ipaddress
import os
import socket
import subprocess
import time

from oslo_config import cfg
from oslo_log import log as logging
import pyroute2

from octavia.amphorae.backends.utils import interface_file
from octavia.common import constants as consts
from octavia.common import exceptions

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class InterfaceController(object):
    ADD = 'add'
    DELETE = 'delete'
    SET = 'set'

    def interface_file_list(self):
        net_dir = interface_file.InterfaceFile.get_directory()

        for f in os.listdir(net_dir):
            for ext in interface_file.InterfaceFile.get_extensions():
                if f.endswith(ext):
                    yield os.path.join(net_dir, f)

    def list(self):
        interfaces = {}
        for f in self.interface_file_list():
            iface = interface_file.InterfaceFile.from_file(f)
            interfaces[iface.name] = iface
        return interfaces

    def _family(self, address):
        return (socket.AF_INET6
                if ipaddress.ip_network(address, strict=False).version == 6
                else socket.AF_INET)

    def _ipr_command(self, method, command,
                     retry_on_invalid_argument=False,
                     retry_interval=.2,
                     raise_on_error=True,
                     max_retries=20,
                     **kwargs):

        for dummy in range(max_retries + 1):
            try:
                method(command, **kwargs)
                break
            except pyroute2.NetlinkError as e:
                if e.code == errno.EINVAL and retry_on_invalid_argument:
                    LOG.debug("Retrying after %d sec.", retry_interval)
                    time.sleep(retry_interval)
                    continue

                if command == self.ADD and e.code != errno.EEXIST:
                    msg = "Cannot call {} {} (with {}): {}".format(
                        method.__name__, command, kwargs, e)
                    if raise_on_error:
                        raise exceptions.AmphoraNetworkConfigException(msg)
                    LOG.error(msg)
                return
        else:
            msg = "Cannot call {} {} (with {}) after {} retries.".format(
                method.__name__, command, kwargs, max_retries)
            if raise_on_error:
                raise exceptions.AmphoraNetworkConfigException(msg)
            LOG.error(msg)

    def _dhclient_up(self, interface_name):
        cmd = ["/sbin/dhclient",
               "-lf",
               "/var/lib/dhclient/dhclient-{}.leases".format(
                   interface_name),
               "-pf",
               "/run/dhclient-{}.pid".format(interface_name),
               interface_name]
        LOG.debug("Running '%s'", cmd)
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def _dhclient_down(self, interface_name):
        cmd = ["/sbin/dhclient",
               "-r",
               "-lf",
               "/var/lib/dhclient/dhclient-{}.leases".format(
                   interface_name),
               "-pf",
               "/run/dhclient-{}.pid".format(interface_name),
               interface_name]
        LOG.debug("Running '%s'", cmd)
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def _ipv6auto_up(self, interface_name):
        # Set values to enable SLAAC on interface_name
        # accept_ra is set to 2 to accept router advertisements if forwarding
        # is enabled on the interface
        for key, value in (('accept_ra', 2),
                           ('autoconf', 1)):
            cmd = ["/sbin/sysctl",
                   "-w",
                   "net.ipv6.conf.{}.{}={}".format(interface_name,
                                                   key, value)]
            LOG.debug("Running '%s'", cmd)
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def _ipv6auto_down(self, interface_name):
        for key, value in (('accept_ra', 0),
                           ('autoconf', 0)):
            cmd = ["/sbin/sysctl",
                   "-w",
                   "net.ipv6.conf.{}.{}={}".format(interface_name,
                                                   key, value)]
            LOG.debug("Running '%s'", cmd)
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)

    def up(self, interface):
        LOG.info("Setting interface %s up", interface.name)

        for address in interface.addresses:
            if address.get(consts.DHCP):
                self._dhclient_up(interface.name)
            if address.get(consts.IPV6AUTO):
                self._ipv6auto_up(interface.name)

        with pyroute2.IPRoute() as ipr:
            idx = ipr.link_lookup(ifname=interface.name)[0]

            self._ipr_command(ipr.link, self.SET, index=idx,
                              state=consts.IFACE_UP, mtu=interface.mtu)

            for address in interface.addresses:
                if (consts.ADDRESS not in address or
                        address.get(consts.DHCP) or
                        address.get(consts.IPV6AUTO)):
                    continue
                address[consts.FAMILY] = self._family(address[consts.ADDRESS])
                LOG.debug("%s: Adding address %s", interface.name, address)
                self._ipr_command(ipr.addr, self.ADD, index=idx, **address)

            for route in interface.routes:
                route[consts.FAMILY] = self._family(route[consts.DST])
                LOG.debug("%s: Adding route %s", interface.name, route)
                # Set retry_on_invalid_argument=True because the interface
                # might not be ready after setting its addresses
                # Note: can we use 'replace' instead of 'add' here?
                # Set raise_on_error to False, possible invalid (user-defined)
                # routes from the subnet's host_routes will not break the
                # script.
                self._ipr_command(ipr.route, self.ADD,
                                  retry_on_invalid_argument=True,
                                  raise_on_error=False,
                                  oif=idx, **route)

            for rule in interface.rules:
                rule[consts.FAMILY] = self._family(rule[consts.SRC])
                LOG.debug("%s: Adding rule %s", interface.name, rule)
                self._ipr_command(ipr.rule, self.ADD,
                                  retry_on_invalid_argument=True,
                                  **rule)

        for script in interface.scripts[consts.IFACE_UP]:
            LOG.debug("%s: Running command '%s'",
                      interface.name, script[consts.COMMAND])
            subprocess.check_output(script[consts.COMMAND].split())

    def down(self, interface):
        LOG.info("Setting interface %s down", interface.name)

        for address in interface.addresses:
            if address.get(consts.DHCP):
                self._dhclient_down(interface.name)
            if address.get(consts.IPV6AUTO):
                self._ipv6auto_down(interface.name)

        with pyroute2.IPRoute() as ipr:
            idx = ipr.link_lookup(ifname=interface.name)[0]

            link = ipr.get_links(idx)[0]
            current_state = link.get(consts.STATE)

            if current_state == consts.IFACE_UP:
                for rule in interface.rules:
                    rule[consts.FAMILY] = self._family(rule[consts.SRC])
                    LOG.debug("%s: Deleting rule %s", interface.name, rule)
                    self._ipr_command(ipr.rule, self.DELETE,
                                      raise_on_error=False, **rule)

                for route in interface.routes:
                    route[consts.FAMILY] = self._family(route[consts.DST])
                    LOG.debug("%s: Deleting route %s", interface.name, route)
                    self._ipr_command(ipr.route, self.DELETE,
                                      raise_on_error=False, oif=idx, **route)

                for address in interface.addresses:
                    if consts.ADDRESS not in address:
                        continue
                    address[consts.FAMILY] = self._family(
                        address[consts.ADDRESS])
                    LOG.debug("%s: Deleting address %s",
                              interface.name, address)
                    self._ipr_command(ipr.addr, self.DELETE,
                                      raise_on_error=False,
                                      index=idx, **address)

                self._ipr_command(ipr.link, self.SET, raise_on_error=False,
                                  index=idx, state=consts.IFACE_DOWN)

        if current_state == consts.IFACE_UP:
            for script in interface.scripts[consts.IFACE_DOWN]:
                LOG.debug("%s: Running command '%s'",
                          interface.name, script[consts.COMMAND])
                try:
                    subprocess.check_output(script[consts.COMMAND].split())
                except Exception as e:
                    LOG.error("Error while running command '%s' on %s: %s",
                              script[consts.COMMAND], interface.name, e)



#    Copyright 2020 Dawson Coleman
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add listener tls versions column

Revision ID: e5493ae5f9a7
Revises: fbd705961c3a
Create Date: 2020-04-19 02:35:28.502424

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'e5493ae5f9a7'
down_revision = 'fbd705961c3a'


def upgrade():
    op.add_column(
        'listener',
        sa.Column('tls_versions', sa.String(512), nullable=True)
    )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

"""update lb and amphora data model for active passive

Revision ID: 357d17a6d5ac
Revises: 298eac0640a7
Create Date: 2015-07-16 17:41:49.029145

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '357d17a6d5ac'
down_revision = '298eac0640a7'


def upgrade():
    op.create_table(
        u'lb_topology',
        sa.Column(u'name', sa.String(36), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'lb_topology',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'SINGLE'},
            {'name': 'ACTIVE_STANDBY'}
        ]
    )

    op.create_table(
        u'amphora_roles',
        sa.Column(u'name', sa.String(36), primary_key=True),
        sa.Column(u'description', sa.String(255), nullable=True)
    )

    insert_table = sql.table(
        u'amphora_roles',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': 'MASTER'},
            {'name': 'BACKUP'},
            {'name': 'STANDALONE'}
        ]
    )

    op.add_column(
        u'load_balancer',
        sa.Column(u'topology', sa.String(36),
                  sa.ForeignKey(u'lb_topology.name',
                                name=u'fk_lb_topology_name'),
                  nullable=True)
    )

    op.add_column(
        u'amphora',
        sa.Column(u'role', sa.String(36),
                  sa.ForeignKey(u'amphora_roles.name',
                                name=u'fk_amphora_roles_name'),
                  nullable=True)
    )



#    Copyright 2019 Red Hat, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def noop_provider_agent(exit_event):
    LOG.info('No-Op provider agent has started.')
    while not exit_event.is_set():
        time.sleep(1)
    LOG.info('No-Op provider agent is exiting.')



#    Copyright 2016 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""adding name column to member and health monitor

Revision ID: 44a2414dd683
Revises: c11292016060
Create Date: 2016-12-19 13:14:58.879793

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '44a2414dd683'
down_revision = 'c11292016060'


tables = ['member', 'health_monitor']


def upgrade():
    for table in tables:
        op.add_column(
            table,
            sa.Column(u'name', sa.String(255), nullable=True)
        )



# Copyright (c) 2014 Rackspace US, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Certificate Generator API
"""
import abc


class CertGenerator(object, metaclass=abc.ABCMeta):
    """Base Cert Generator Interface

    A Certificate Generator is responsible for generating private keys,
    generating CSRs, and signing TLS certificates.
    """

    @abc.abstractmethod
    def sign_cert(self, csr, validity):
        """Generates a signed certificate from the provided CSR

        This call is designed to block until a signed certificate can be
        returned.

        :param csr: A Certificate Signing Request
        :param validity: Valid for <validity> seconds from the current time

        :return: PEM Encoded Signed certificate
        :raises Exception: If certificate signing fails
        """

    @abc.abstractmethod
    def generate_cert_key_pair(self, cn, validity, bit_length, passphrase):
        """Generates a private key and certificate pair

        :param cn: Common name to use for the Certificate
        :param validity: Validity period for the Certificate
        :param bit_length: Private key bit length
        :param passphrase: Passphrase to use for encrypting the private key

        :return: octavia.certificates.common.Cert representation of the
                 certificate data
        :raises Exception: If generation fails
        """



# Copyright 2018 Huawei
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Extend the l7rule type for support client certificate cases

Revision ID: 1afc932f1ca2
Revises: ffad172e98c1
Create Date: 2018-10-03 20:47:52.405865

"""


from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '1afc932f1ca2'
down_revision = 'ffad172e98c1'

new_fields = ['SSL_CONN_HAS_CERT', 'SSL_VERIFY_RESULT', 'SSL_DN_FIELD']


def upgrade():

    insert_table = sql.table(
        u'l7rule_type',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )
    cows = [{'name': field} for field in new_fields]
    op.bulk_insert(insert_table, cows)



# Copyright 2017 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import subprocess

import distro
from oslo_config import cfg
from oslo_log import log as logging
import webob
from werkzeug import exceptions

from octavia.amphorae.backends.utils import interface_file
from octavia.common import constants as consts
from octavia.common import exceptions as octavia_exceptions

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class BaseOS(object):

    def __init__(self, os_name):
        self.os_name = os_name
        self.package_name_map = {}

    @classmethod
    def _get_subclasses(cls):
        for subclass in cls.__subclasses__():
            for sc in subclass._get_subclasses():
                yield sc
            yield subclass

    @classmethod
    def get_os_util(cls):
        os_name = distro.id()
        for subclass in cls._get_subclasses():
            if subclass.is_os_name(os_name):
                return subclass(os_name)
        raise octavia_exceptions.InvalidAmphoraOperatingSystem(os_name=os_name)

    def _map_package_name(self, package_name):
        return self.package_name_map.get(package_name, package_name)

    def write_interface_file(self, interface, ip_address, prefixlen):
        interface = interface_file.InterfaceFile(
            name=interface,
            addresses=[{
                "address": ip_address,
                "prefixlen": prefixlen
            }]
        )
        interface.write()

    def write_vip_interface_file(self, interface, vip, ip_version,
                                 prefixlen, gateway,
                                 mtu, vrrp_ip,
                                 host_routes):
        vip_interface = interface_file.VIPInterfaceFile(
            name=interface,
            mtu=mtu,
            vip=vip,
            ip_version=ip_version,
            prefixlen=prefixlen,
            gateway=gateway,
            vrrp_ip=vrrp_ip,
            host_routes=host_routes,
            topology=CONF.controller_worker.loadbalancer_topology)
        vip_interface.write()

    def write_port_interface_file(self, interface, fixed_ips, mtu):
        port_interface = interface_file.PortInterfaceFile(
            name=interface,
            mtu=mtu,
            fixed_ips=fixed_ips)
        port_interface.write()

    @classmethod
    def _bring_if_up(cls, interface, what):
        cmd = ("ip netns exec {ns} amphora-interface up {params}".format(
            ns=consts.AMPHORA_NAMESPACE, params=interface))
        LOG.debug("Executing: %s", cmd)
        try:
            out = subprocess.check_output(cmd.split(),
                                          stderr=subprocess.STDOUT)
            LOG.debug(out)
        except subprocess.CalledProcessError as e:
            LOG.error('Failed to set up %s due to error: %s %s', interface,
                      e, e.output)
            raise exceptions.HTTPException(
                response=webob.Response(json=dict(
                    message='Error plugging {0}'.format(what),
                    details=e.output), status=500))

    @classmethod
    def _bring_if_down(cls, interface):
        cmd = ("ip netns exec {ns} amphora-interface down {params}".format(
            ns=consts.AMPHORA_NAMESPACE, params=interface))
        LOG.debug("Executing: %s", cmd)
        try:
            subprocess.check_output(cmd.split(), stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            LOG.info('Ignoring failure to set %s down due to error: %s %s',
                     interface, e, e.output)

    @classmethod
    def bring_interfaces_up(cls, ip, primary_interface):
        cls._bring_if_down(primary_interface)
        cls._bring_if_up(primary_interface, 'VIP')


class Ubuntu(BaseOS):

    @classmethod
    def is_os_name(cls, os_name):
        return os_name in ['ubuntu']

    def cmd_get_version_of_installed_package(self, package_name):
        name = self._map_package_name(package_name)
        return "dpkg-query -W -f=${{Version}} {name}".format(name=name)


class RH(BaseOS):

    @classmethod
    def is_os_name(cls, os_name):
        return os_name in ['fedora', 'rhel']

    def cmd_get_version_of_installed_package(self, package_name):
        name = self._map_package_name(package_name)
        return "rpm -q --queryformat %{{VERSION}} {name}".format(name=name)


class CentOS(RH):

    def __init__(self, os_name):
        super().__init__(os_name)
        if distro.version() == '7':
            self.package_name_map.update({'haproxy': 'haproxy18'})

    @classmethod
    def is_os_name(cls, os_name):
        return os_name in ['centos']



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from wsme import exc
from wsme.rest import json as wsme_json

from octavia.api.common import types as base_type
from octavia.common import constants
from octavia.tests.unit import base


def build_body(mandatory_fields, extra_attributes):
    body = {}
    for key in mandatory_fields:
        body[key] = mandatory_fields[key]
    for key in extra_attributes:
        body[key] = extra_attributes[key]
    return body


class BaseTypesTest(base.TestCase):
    _type = base_type.BaseType
    _mandatory_fields = {}


class BaseTestUuid(base.TestCase):

    def assert_uuid_attr(self, attr):
        kwargs = {attr: uuidutils.generate_uuid()}
        self._type(**kwargs)

    def assert_uuid_attr_fail_with_integer(self, attr):
        kwargs = {attr: 1}
        self.assertRaises(exc.InvalidInput, self._type, **kwargs)

    def assert_uuid_attr_fail_with_short_str(self, attr):
        kwargs = {attr: '12345'}
        self.assertRaises(exc.InvalidInput, self._type, **kwargs)

    def assert_uuid_attr_fail_with_shorter_than_uuid(self, attr):
        kwargs = {attr: uuidutils.generate_uuid()[1:]}
        self.assertRaises(exc.InvalidInput, self._type, **kwargs)

    def assert_uuid_attr_fail_with_longer_than_uuid(self, attr):
        kwargs = {attr: uuidutils.generate_uuid() + "0"}
        self.assertRaises(exc.InvalidInput, self._type, **kwargs)


class BaseTestString(base.TestCase):

    def _default_min_max_lengths(self, min_length=None, max_length=None):
        if max_length is None:
            if min_length is None:
                max_length = 255
                min_length = 2
            else:
                max_length = min_length + 1
        else:
            if min_length is None:
                min_length = max_length - 1
        return min_length, max_length

    def assert_string_attr(self, attr, min_length=None, max_length=None):
        min_length, max_length = self._default_min_max_lengths(min_length,
                                                               max_length)
        string_val = 'a' * (max_length - 1)
        kwargs = {attr: string_val}
        self._type(**kwargs)

    def assert_string_attr_min_length(self, attr, min_length):
        min_length, max_length = self._default_min_max_lengths(min_length)
        string_val = 'a' * (min_length - 1)
        kwargs = {attr: string_val}
        # No point in testing if min_length is <= 0
        if min_length > 0:
            self.assertRaises(exc.InvalidInput, self._type, **kwargs)

    def assert_string_attr_max_length(self, attr, max_length=None):
        min_length, max_length = self._default_min_max_lengths(max_length)
        string_val = 'a' * (max_length + 1)
        kwargs = {attr: string_val}
        self.assertRaises(exc.InvalidInput, self._type, **kwargs)


class BaseTestBool(base.TestCase):

    def assert_bool_attr(self, attr):
        kwargs = {attr: True}
        self.assertIsNotNone(self._type(**kwargs))
        kwargs = {attr: False}
        self.assertIsNotNone(self._type(**kwargs))

    def assert_bool_attr_non_bool(self, attr):
        kwargs = {attr: 'test'}
        self.assertRaises(exc.InvalidInput, self._type, **kwargs)


class TestIdMixin(BaseTestUuid):
    id_attr = 'id'

    def test_id(self):
        self.assert_uuid_attr(self.id_attr)
        self.assert_uuid_attr_fail_with_integer(self.id_attr)
        self.assert_uuid_attr_fail_with_short_str(self.id_attr)
        self.assert_uuid_attr_fail_with_shorter_than_uuid(self.id_attr)
        self.assert_uuid_attr_fail_with_longer_than_uuid(self.id_attr)

    def test_id_readonly(self):
        body = build_body(self._mandatory_fields,
                          {self.id_attr: uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)


class TestProjectIdMixin(BaseTestUuid):
    project_id_attr = 'project_id'

    def test_project_id(self):
        self.assert_uuid_attr(self.project_id_attr)
        self.assert_uuid_attr_fail_with_integer(self.project_id_attr)
        self.assert_uuid_attr_fail_with_short_str(self.project_id_attr)
        self.assert_uuid_attr_fail_with_shorter_than_uuid(self.project_id_attr)
        self.assert_uuid_attr_fail_with_longer_than_uuid(self.project_id_attr)

    def test_project_id_readonly(self):
        body = build_body(self._mandatory_fields,
                          {self.project_id_attr: uuidutils.generate_uuid()})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)


class TestNameMixin(BaseTestString):
    name_attr = 'name'

    def test_name(self):
        self.assert_string_attr(self.name_attr, min_length=0, max_length=255)
        self.assert_string_attr_min_length(self.name_attr, 0)
        self.assert_string_attr_max_length(self.name_attr, 255)

    def test_editable_name(self):
        name = "Name"
        body = build_body(self._mandatory_fields, {self.name_attr: name})
        type_instance = wsme_json.fromjson(self._type, body)
        self.assertEqual(name, type_instance.name)


class TestDescriptionMixin(BaseTestString):
    description_attr = 'description'

    def test_description(self):
        self.assert_string_attr(self.description_attr, min_length=0,
                                max_length=255)
        self.assert_string_attr_min_length(self.description_attr, 0)
        self.assert_string_attr_max_length(self.description_attr, 255)

    def test_editable_description(self):
        description = "Description"
        body = build_body(self._mandatory_fields,
                          {self.description_attr: description})
        type_instance = wsme_json.fromjson(self._type, body)
        self.assertEqual(description, type_instance.description)


class TestEnabledMixin(BaseTestBool):
    enabled_attr = 'enabled'

    def test_enabled(self):
        self.assert_bool_attr(self.enabled_attr)
        self.assert_bool_attr_non_bool(self.enabled_attr)

    def test_default_enabled_true(self):
        body = build_body(self._mandatory_fields, {})
        type_instance = wsme_json.fromjson(self._type, body)
        self.assertTrue(type_instance.enabled)

    def test_editable_enabled(self):
        body = build_body(self._mandatory_fields, {"enabled": False})
        type_instance = wsme_json.fromjson(self._type, body)
        self.assertFalse(type_instance.enabled)


class TestProvisioningStatusMixin(BaseTestString):
    provisioning_attr = 'provisioning_status'

    def test_provisioning_status(self):
        self.assert_string_attr(self.provisioning_attr, min_length=0,
                                max_length=16)
        self.assert_string_attr_min_length(self.provisioning_attr, 0)
        self.assert_string_attr_max_length(self.provisioning_attr, 16)

    def test_provisioning_status_readonly(self):
        status = constants.ACTIVE
        body = build_body(self._mandatory_fields,
                          {self.provisioning_attr: status})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)


class TestOperatingStatusMixin(BaseTestString):
    operating_attr = 'operating_status'

    def test_operating_status(self):
        self.assert_string_attr(self.operating_attr, min_length=0,
                                max_length=16)
        self.assert_string_attr_min_length(self.operating_attr, 0)
        self.assert_string_attr_max_length(self.operating_attr, 16)

    def test_operating_status_readonly(self):
        status = constants.ONLINE
        body = build_body(self._mandatory_fields,
                          {self.operating_attr: status})
        self.assertRaises(exc.InvalidInput, wsme_json.fromjson,
                          self._type, body)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Adding TERMINATED_HTTPS support and TLS ref ID char length increase

Revision ID: 2351ea316465
Revises: 48660b6643f0
Create Date: 2015-05-22 11:57:04.703910

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy import sql

# revision identifiers, used by Alembic.
revision = '2351ea316465'
down_revision = '357d17a6d5ac'


new_protocol = 'TERMINATED_HTTPS'


def upgrade():
    insert_table = sql.table(
        u'protocol',
        sql.column(u'name', sa.String),
        sql.column(u'description', sa.String)
    )

    op.bulk_insert(
        insert_table,
        [
            {'name': new_protocol}
        ]
    )
    op.alter_column(u'listener', u'tls_certificate_id',
                    existing_type=sa.String(255), nullable=True)



# Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import subprocess
from unittest import mock

import flask

from octavia.amphorae.backends.agent.api_server import keepalived
import octavia.tests.unit.base as base


class KeepalivedTestCase(base.TestCase):
    def setUp(self):
        super().setUp()
        self.app = flask.Flask(__name__)
        self.client = self.app.test_client()
        self._ctx = self.app.test_request_context()
        self._ctx.push()
        self.test_keepalived = keepalived.Keepalived()

    @mock.patch('subprocess.check_output')
    def test_manager_keepalived_service(self, mock_check_output):
        res = self.test_keepalived.manager_keepalived_service('start')
        cmd = ("/usr/sbin/service octavia-keepalived {action}".format(
            action='start'))
        mock_check_output.assert_called_once_with(cmd.split(),
                                                  stderr=subprocess.STDOUT)
        self.assertEqual(202, res.status_code)

        res = self.test_keepalived.manager_keepalived_service('restart')
        self.assertEqual(400, res.status_code)

        mock_check_output.side_effect = subprocess.CalledProcessError(1,
                                                                      'blah!')

        res = self.test_keepalived.manager_keepalived_service('start')
        self.assertEqual(500, res.status_code)



# Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import jinja2

from octavia.common.config import cfg
from octavia.common import constants

CONF = cfg.CONF

TEMPLATES_DIR = (os.path.dirname(os.path.realpath(__file__)) +
                 constants.LOGGING_TEMPLATES + '/')


class LoggingJinjaTemplater(object):

    def __init__(self, logging_templates=None):
        self.logging_templates = logging_templates or TEMPLATES_DIR
        template_loader = jinja2.FileSystemLoader(searchpath=os.path.dirname(
            self.logging_templates))
        jinja_env = jinja2.Environment(loader=template_loader, autoescape=True)
        self.logging_template = jinja_env.get_template(
            constants.LOGGING_CONF_TEMPLATE)

    def build_logging_config(self):
        admin_log_hosts = []
        for server in CONF.amphora_agent.admin_log_targets or []:
            (host, port) = server.rsplit(':', 1)
            admin_log_hosts.append({
                'host': host,
                'port': port,
            })
        tenant_log_hosts = []
        for server in CONF.amphora_agent.tenant_log_targets or []:
            (host, port) = server.rsplit(':', 1)
            tenant_log_hosts.append({
                'host': host,
                'port': port,
            })
        return self.logging_template.render(
            {'admin_log_hosts': admin_log_hosts,
             'tenant_log_hosts': tenant_log_hosts,
             'protocol': CONF.amphora_agent.log_protocol,
             'retry_count': CONF.amphora_agent.log_retry_count,
             'retry_interval': CONF.amphora_agent.log_retry_interval,
             'queue_size': CONF.amphora_agent.log_queue_size,
             'forward_all_logs': CONF.amphora_agent.forward_all_logs,
             'disable_local_log_storage':
                 CONF.amphora_agent.disable_local_log_storage,
             'admin_log_facility':
                 CONF.amphora_agent.administrative_log_facility,
             'user_log_facility': CONF.amphora_agent.user_log_facility,
             })



# Copyright (c) 2014 Rackspace US, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Barbican ACL auth class for Barbican certificate handling
"""
from barbicanclient import client as barbican_client
from keystoneauth1.identity.generic import token
from keystoneauth1 import session

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from octavia.certificates.common import barbican as barbican_common
from octavia.common import keystone

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class BarbicanACLAuth(barbican_common.BarbicanAuth):
    _barbican_client = None

    @classmethod
    def get_barbican_client(cls, project_id=None):
        if not cls._barbican_client:
            try:
                ksession = keystone.KeystoneSession()
                cls._barbican_client = barbican_client.Client(
                    session=ksession.get_session(),
                    region_name=CONF.certificates.region_name,
                    interface=CONF.certificates.endpoint_type
                )
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.exception("Error creating Barbican client")
        return cls._barbican_client

    @classmethod
    def ensure_secret_access(cls, context, ref):
        # get a normal session
        ksession = keystone.KeystoneSession()
        user_id = ksession.get_service_user_id()

        # use barbican client to set the ACLs
        bc = cls.get_barbican_client_user_auth(context)
        acl = bc.acls.get(ref)
        read_oper = acl.get('read')
        if user_id not in read_oper.users:
            read_oper.users.append(user_id)
            acl.submit()

    @classmethod
    def revoke_secret_access(cls, context, ref):
        # get a normal session
        ksession = keystone.KeystoneSession()
        user_id = ksession.get_service_user_id()

        # use barbican client to set the ACLs
        bc = cls.get_barbican_client_user_auth(context)
        acl = bc.acls.get(ref)
        read_oper = acl.get('read')
        if user_id in read_oper.users:
            read_oper.users.remove(user_id)
            acl.submit()

    @classmethod
    def get_barbican_client_user_auth(cls, context):
        # get a normal session
        ksession = keystone.KeystoneSession()
        service_auth = ksession.get_auth()

        # make our own auth and swap it in
        user_auth = token.Token(auth_url=service_auth.auth_url,
                                token=context.auth_token,
                                project_id=context.project_id)
        user_session = session.Session(
            auth=user_auth,
            verify=CONF.certificates.ca_certificates_file)

        # create a special barbican client with our user's session
        return barbican_client.Client(
            session=user_session,
            region_name=CONF.certificates.region_name,
            interface=CONF.certificates.endpoint_type)



#    Copyright 2019 Verizon Media
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
from oslo_db import exception as odb_exceptions
from oslo_utils import uuidutils

from octavia.common import constants
import octavia.common.context
from octavia.tests.functional.api.v2 import base


class TestAvailabilityZoneProfiles(base.BaseAPITest):
    root_tag = 'availability_zone_profile'
    root_tag_list = 'availability_zone_profiles'
    root_tag_links = 'availability_zone_profile_links'

    def _assert_request_matches_response(self, req, resp, **optionals):
        self.assertTrue(uuidutils.is_uuid_like(resp.get('id')))
        self.assertEqual(req.get('name'), resp.get('name'))
        self.assertEqual(req.get(constants.PROVIDER_NAME),
                         resp.get(constants.PROVIDER_NAME))
        self.assertEqual(req.get(constants.AVAILABILITY_ZONE_DATA),
                         resp.get(constants.AVAILABILITY_ZONE_DATA))

    def test_empty_list(self):
        response = self.get(self.AZPS_PATH)
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual([], api_list)

    def test_create(self):
        az_json = {'name': 'test1', constants.PROVIDER_NAME: 'noop_driver',
                   constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(az_json)
        response = self.post(self.AZPS_PATH, body)
        api_azp = response.json.get(self.root_tag)
        self._assert_request_matches_response(az_json, api_azp)

    def test_create_with_missing_name(self):
        az_json = {constants.PROVIDER_NAME: 'pr1',
                   constants.AVAILABILITY_ZONE_DATA: '{"x": "y"}'}
        body = self._build_body(az_json)
        response = self.post(self.AZPS_PATH, body, status=400)
        err_msg = ("Invalid input for field/attribute name. Value: "
                   "'None'. Mandatory field missing.")
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_with_missing_provider(self):
        az_json = {'name': 'xyz',
                   constants.AVAILABILITY_ZONE_DATA: '{"x": "y"}'}
        body = self._build_body(az_json)
        response = self.post(self.AZPS_PATH, body, status=400)
        err_msg = ("Invalid input for field/attribute provider_name. "
                   "Value: 'None'. Mandatory field missing.")
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_with_missing_availability_zone_data(self):
        az_json = {'name': 'xyz', constants.PROVIDER_NAME: 'pr1'}
        body = self._build_body(az_json)
        response = self.post(self.AZPS_PATH, body, status=400)
        err_msg = ("Invalid input for field/attribute availability_zone_data. "
                   "Value: 'None'. Mandatory field missing.")
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_create_with_empty_availability_zone_data(self):
        az_json = {'name': 'test1', constants.PROVIDER_NAME: 'noop_driver',
                   constants.AVAILABILITY_ZONE_DATA: '{}'}
        body = self._build_body(az_json)
        response = self.post(self.AZPS_PATH, body)
        api_azp = response.json.get(self.root_tag)
        self._assert_request_matches_response(az_json, api_azp)

    def test_create_with_long_name(self):
        az_json = {'name': 'n' * 256, constants.PROVIDER_NAME: 'test1',
                   constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(az_json)
        self.post(self.AZPS_PATH, body, status=400)

    def test_create_with_long_provider(self):
        az_json = {'name': 'name1', constants.PROVIDER_NAME: 'n' * 256,
                   constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(az_json)
        self.post(self.AZPS_PATH, body, status=400)

    def test_create_with_long_availability_zone_data(self):
        az_json = {'name': 'name1', constants.PROVIDER_NAME: 'amp',
                   constants.AVAILABILITY_ZONE_DATA: 'n' * 4097}
        body = self._build_body(az_json)
        self.post(self.AZPS_PATH, body, status=400)

    def test_create_authorized(self):
        az_json = {'name': 'test1', constants.PROVIDER_NAME: 'noop_driver',
                   constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(az_json)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.post(self.AZPS_PATH, body)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        api_azp = response.json.get(self.root_tag)
        self._assert_request_matches_response(az_json, api_azp)

    def test_create_not_authorized(self):
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        az_json = {'name': 'name',
                   constants.PROVIDER_NAME: 'xyz',
                   constants.AVAILABILITY_ZONE_DATA: '{"x": "y"}'}
        body = self._build_body(az_json)
        response = self.post(self.AZPS_PATH, body, status=403)
        api_azp = response.json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, api_azp)

    def test_create_db_failure(self):
        az_json = {'name': 'test1', constants.PROVIDER_NAME: 'noop_driver',
                   constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(az_json)
        with mock.patch(
                "octavia.db.repositories.AvailabilityZoneProfileRepository."
                "create") as mock_create:
            mock_create.side_effect = Exception
            self.post(self.AZPS_PATH, body, status=500)

            mock_create.side_effect = odb_exceptions.DBDuplicateEntry
            self.post(self.AZPS_PATH, body, status=409)

    def test_create_with_invalid_json(self):
        az_json = {'name': 'test1', constants.PROVIDER_NAME: 'noop_driver',
                   constants.AVAILABILITY_ZONE_DATA: '{hello: "world"}'}
        body = self._build_body(az_json)
        self.post(self.AZPS_PATH, body, status=400)

    def test_get(self):
        azp = self.create_availability_zone_profile(
            'name', 'noop_driver', '{"x": "y"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        response = self.get(
            self.AZP_PATH.format(
                azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('name', response.get('name'))
        self.assertEqual(azp.get('id'), response.get('id'))

    def test_get_one_deleted_id(self):
        response = self.get(self.AZP_PATH.format(azp_id=constants.NIL_UUID),
                            status=404)
        self.assertEqual('Availability Zone Profile {} not found.'.format(
            constants.NIL_UUID), response.json.get('faultstring'))

    def test_get_one_fields_filter(self):
        azp = self.create_availability_zone_profile(
            'name', 'noop_driver', '{"x": "y"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id')), params={
                'fields': ['id', constants.PROVIDER_NAME]}
        ).json.get(self.root_tag)
        self.assertEqual(azp.get('id'), response.get('id'))
        self.assertIn(u'id', response)
        self.assertIn(constants.PROVIDER_NAME, response)
        self.assertNotIn(u'name', response)
        self.assertNotIn(constants.AVAILABILITY_ZONE_DATA, response)

    def test_get_authorized(self):
        azp = self.create_availability_zone_profile(
            'name', 'noop_driver', '{"x": "y"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.get(
                    self.AZP_PATH.format(
                        azp_id=azp.get('id'))).json.get(self.root_tag)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual('name', response.get('name'))
        self.assertEqual(azp.get('id'), response.get('id'))

    def test_get_not_authorized(self):
        azp = self.create_availability_zone_profile(
            'name', 'noop_driver', '{"x": "y"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        self.get(self.AZP_PATH.format(azp_id=azp.get('id')), status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

    def test_get_all(self):
        fp1 = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        ref_fp_1 = {u'availability_zone_data': u'{"compute_zone": "my_az_1"}',
                    u'id': fp1.get('id'), u'name': u'test1',
                    constants.PROVIDER_NAME: u'noop_driver'}
        self.assertTrue(uuidutils.is_uuid_like(fp1.get('id')))
        fp2 = self.create_availability_zone_profile(
            'test2', 'noop_driver-alt', '{"compute_zone": "my_az_1"}')
        ref_fp_2 = {u'availability_zone_data': u'{"compute_zone": "my_az_1"}',
                    u'id': fp2.get('id'), u'name': u'test2',
                    constants.PROVIDER_NAME: u'noop_driver-alt'}
        self.assertTrue(uuidutils.is_uuid_like(fp2.get('id')))

        response = self.get(self.AZPS_PATH)
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual(2, len(api_list))
        self.assertIn(ref_fp_1, api_list)
        self.assertIn(ref_fp_2, api_list)

    def test_get_all_fields_filter(self):
        fp1 = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(fp1.get('id')))
        fp2 = self.create_availability_zone_profile(
            'test2', 'noop_driver-alt', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(fp2.get('id')))

        response = self.get(self.AZPS_PATH, params={
            'fields': ['id', 'name']})
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual(2, len(api_list))
        for profile in api_list:
            self.assertIn(u'id', profile)
            self.assertIn(u'name', profile)
            self.assertNotIn(constants.PROVIDER_NAME, profile)
            self.assertNotIn(constants.AVAILABILITY_ZONE_DATA, profile)

    def test_get_all_authorized(self):
        fp1 = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(fp1.get('id')))
        fp2 = self.create_availability_zone_profile(
            'test2', 'noop_driver-alt', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(fp2.get('id')))
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.get(self.AZPS_PATH)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        api_list = response.json.get(self.root_tag_list)
        self.assertEqual(2, len(api_list))

    def test_get_all_not_authorized(self):
        fp1 = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(fp1.get('id')))
        fp2 = self.create_availability_zone_profile(
            'test2', 'noop_driver-alt', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(fp2.get('id')))

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        self.get(self.AZPS_PATH, status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)

    def test_update(self):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        update_data = {'name': 'the_profile',
                       constants.PROVIDER_NAME: 'noop_driver-alt',
                       constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(update_data)
        self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('the_profile', response.get('name'))
        self.assertEqual('noop_driver-alt',
                         response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"hello": "world"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

    def test_update_deleted_id(self):
        update_data = {'name': 'fake_profile'}
        body = self._build_body(update_data)
        response = self.put(self.AZP_PATH.format(azp_id=constants.NIL_UUID),
                            body, status=404)
        self.assertEqual('Availability Zone Profile {} not found.'.format(
            constants.NIL_UUID), response.json.get('faultstring'))

    def test_update_nothing(self):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        body = self._build_body({})
        self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('test_profile', response.get('name'))
        self.assertEqual('noop_driver', response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"x": "y"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

    def test_update_name_none(self):
        self._test_update_param_none(constants.NAME)

    def test_update_provider_name_none(self):
        self._test_update_param_none(constants.PROVIDER_NAME)

    def test_update_availability_zone_data_none(self):
        self._test_update_param_none(constants.AVAILABILITY_ZONE_DATA)

    def _test_update_param_none(self, param_name):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        expect_error_msg = ("None is not a valid option for %s" %
                            param_name)
        body = self._build_body({param_name: None})
        response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body,
                            status=400)
        self.assertEqual(expect_error_msg, response.json['faultstring'])

    def test_update_no_availability_zone_data(self):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        update_data = {'name': 'the_profile',
                       constants.PROVIDER_NAME: 'noop_driver-alt'}
        body = self._build_body(update_data)
        response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('the_profile', response.get('name'))
        self.assertEqual('noop_driver-alt',
                         response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"x": "y"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

    def test_update_authorized(self):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        update_data = {'name': 'the_profile',
                       constants.PROVIDER_NAME: 'noop_driver-alt',
                       constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(update_data)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')),
                                    body)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('the_profile', response.get('name'))
        self.assertEqual('noop_driver-alt',
                         response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"hello": "world"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

    def test_update_not_authorized(self):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        update_data = {'name': 'the_profile', constants.PROVIDER_NAME: 'amp',
                       constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(update_data)
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')),
                            body, status=403)
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('test_profile', response.get('name'))
        self.assertEqual('noop_driver', response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"x": "y"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

    def test_update_in_use(self):
        azp = self.create_availability_zone_profile(
            'test_profile', 'noop_driver', '{"x": "y"}')
        self.create_availability_zone(
            'name1', 'description', azp.get('id'), True)

        # Test updating provider while in use is not allowed
        update_data = {'name': 'the_profile',
                       constants.PROVIDER_NAME: 'noop_driver-alt'}
        body = self._build_body(update_data)
        response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body,
                            status=409)
        err_msg = ("Availability Zone Profile {} is in use and cannot be "
                   "modified.".format(azp.get('id')))
        self.assertEqual(err_msg, response.json.get('faultstring'))
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('test_profile', response.get('name'))
        self.assertEqual('noop_driver', response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"x": "y"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

        # Test updating availability zone data while in use is not allowed
        update_data = {'name': 'the_profile',
                       constants.AVAILABILITY_ZONE_DATA: '{"hello": "world"}'}
        body = self._build_body(update_data)
        response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body,
                            status=409)
        err_msg = ("Availability Zone Profile {} is in use and cannot be "
                   "modified.".format(azp.get('id')))
        self.assertEqual(err_msg, response.json.get('faultstring'))
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('test_profile', response.get('name'))
        self.assertEqual('noop_driver', response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"x": "y"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

        # Test that you can still update the name when in use
        update_data = {'name': 'the_profile'}
        body = self._build_body(update_data)
        response = self.put(self.AZP_PATH.format(azp_id=azp.get('id')), body)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('the_profile', response.get('name'))
        self.assertEqual('noop_driver', response.get(constants.PROVIDER_NAME))
        self.assertEqual('{"x": "y"}',
                         response.get(constants.AVAILABILITY_ZONE_DATA))

    def test_delete(self):
        azp = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        self.delete(self.AZP_PATH.format(azp_id=azp.get('id')))
        response = self.get(self.AZP_PATH.format(
            azp_id=azp.get('id')), status=404)
        err_msg = "Availability Zone Profile %s not found." % azp.get('id')
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_delete_deleted_id(self):
        response = self.delete(self.AZP_PATH.format(azp_id=constants.NIL_UUID),
                               status=404)
        self.assertEqual('Availability Zone Profile {} not found.'.format(
            constants.NIL_UUID), response.json.get('faultstring'))

    def test_delete_nonexistent_id(self):
        response = self.delete(self.AZP_PATH.format(azp_id='bogus_id'),
                               status=404)
        self.assertEqual('Availability Zone Profile bogus_id not found.',
                         response.json.get('faultstring'))

    def test_delete_authorized(self):
        azp = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)
        project_id = uuidutils.generate_uuid()
        with mock.patch.object(octavia.common.context.Context, 'project_id',
                               project_id):
            override_credentials = {
                'service_user_id': None,
                'user_domain_id': None,
                'is_admin_project': True,
                'service_project_domain_id': None,
                'service_project_id': None,
                'roles': ['load-balancer_member'],
                'user_id': None,
                'is_admin': True,
                'service_user_domain_id': None,
                'project_domain_id': None,
                'service_roles': [],
                'project_id': project_id}
            with mock.patch(
                    "oslo_context.context.RequestContext.to_policy_values",
                    return_value=override_credentials):
                self.delete(self.AZP_PATH.format(azp_id=azp.get('id')))
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        response = self.get(self.AZP_PATH.format(
            azp_id=azp.get('id')), status=404)
        err_msg = "Availability Zone Profile %s not found." % azp.get('id')
        self.assertEqual(err_msg, response.json.get('faultstring'))

    def test_delete_not_authorized(self):
        azp = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.assertTrue(uuidutils.is_uuid_like(azp.get('id')))
        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        auth_strategy = self.conf.conf.api_settings.get('auth_strategy')
        self.conf.config(group='api_settings', auth_strategy=constants.TESTING)

        response = self.delete(self.AZP_PATH.format(
            azp_id=azp.get('id')), status=403)
        api_azp = response.json
        self.conf.config(group='api_settings', auth_strategy=auth_strategy)
        self.assertEqual(self.NOT_AUTHORIZED_BODY, api_azp)
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('test1', response.get('name'))

    def test_delete_in_use(self):
        azp = self.create_availability_zone_profile(
            'test1', 'noop_driver', '{"compute_zone": "my_az_1"}')
        self.create_availability_zone(
            'name1', 'description', azp.get('id'), True)
        response = self.delete(self.AZP_PATH.format(azp_id=azp.get('id')),
                               status=409)
        err_msg = ("Availability Zone Profile {} is in use and cannot be "
                   "modified.".format(azp.get('id')))
        self.assertEqual(err_msg, response.json.get('faultstring'))
        response = self.get(
            self.AZP_PATH.format(azp_id=azp.get('id'))).json.get(self.root_tag)
        self.assertEqual('test1', response.get('name'))



# Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from octavia_lib.api.drivers import exceptions as driver_exceptions
from octavia_lib.common import constants as lib_consts
from oslo_log import log as logging
from oslo_utils import excutils

from octavia.common import constants as consts
from octavia.common import data_models
from octavia.common import utils
from octavia.db import api as db_apis
from octavia.db import repositories as repo
from octavia.statistics import stats_base

LOG = logging.getLogger(__name__)


class DriverUpdater(object):

    def __init__(self, **kwargs):
        self.repos = repo.Repositories()
        self.loadbalancer_repo = repo.LoadBalancerRepository()
        self.listener_repo = repo.ListenerRepository()
        self.pool_repo = repo.PoolRepository()
        self.health_mon_repo = repo.HealthMonitorRepository()
        self.member_repo = repo.MemberRepository()
        self.l7policy_repo = repo.L7PolicyRepository()
        self.l7rule_repo = repo.L7RuleRepository()
        self.listener_stats_repo = repo.ListenerStatisticsRepository()

        self.db_session = db_apis.get_session()
        super().__init__(**kwargs)

    def _check_for_lb_vip_deallocate(self, repo, lb_id):
        lb = repo.get(self.db_session, id=lb_id)
        if lb.vip.octavia_owned:
            vip = lb.vip
            # We need a backreference
            vip.load_balancer = lb
            # Only lookup the network driver if we have a VIP to deallocate
            network_driver = utils.get_network_driver()
            network_driver.deallocate_vip(vip)

    def _decrement_quota(self, repo, object_name, record_id):
        lock_session = db_apis.get_session(autocommit=False)
        db_object = repo.get(lock_session, id=record_id)
        try:
            if db_object.provisioning_status == consts.DELETED:
                LOG.info('%(name)s with ID of %(id)s is already in the '
                         'DELETED state. Skipping quota update.',
                         {'name': object_name, 'id': record_id})
                lock_session.rollback()
                return
            self.repos.decrement_quota(lock_session,
                                       repo.model_class.__data_model__,
                                       db_object.project_id)
            lock_session.commit()
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error('Failed to decrement %(name)s quota for '
                          'project: %(proj)s the project may have excess '
                          'quota in use.', {'proj': db_object.project_id,
                                            'name': object_name})
                lock_session.rollback()

    def _process_status_update(self, repo, object_name, record,
                               delete_record=False):
        # Zero it out so that if the ID is missing from a record we do not
        # report the last LB as the failed record in the exception
        record_id = None
        try:
            record_id = record['id']
            record_kwargs = {}
            prov_status = record.get(consts.PROVISIONING_STATUS, None)
            if prov_status:
                if prov_status == consts.DELETED:
                    if object_name == consts.LOADBALANCERS:
                        self._check_for_lb_vip_deallocate(repo, record_id)

                    self._decrement_quota(repo, object_name, record_id)

                    if delete_record and object_name != consts.LOADBALANCERS:
                        repo.delete(self.db_session, id=record_id)
                        return

                record_kwargs[consts.PROVISIONING_STATUS] = prov_status
            op_status = record.get(consts.OPERATING_STATUS, None)
            if op_status:
                record_kwargs[consts.OPERATING_STATUS] = op_status
            if prov_status or op_status:
                repo.update(self.db_session, record_id, **record_kwargs)
        except Exception as e:
            # We need to raise a failure here to notify the driver it is
            # sending bad status data.
            raise driver_exceptions.UpdateStatusError(
                fault_string=str(e), status_object_id=record_id,
                status_object=object_name)

    def update_loadbalancer_status(self, status):
        """Update load balancer status.

        :param status: dictionary defining the provisioning status and
            operating status for load balancer objects, including pools,
            members, listeners, L7 policies, and L7 rules.
            iod (string): ID for the object.
            provisioning_status (string): Provisioning status for the object.
            operating_status (string): Operating status for the object.
        :type status: dict
        :raises: UpdateStatusError
        :returns: None
        """
        try:
            members = status.pop(consts.MEMBERS, [])
            for member in members:
                self._process_status_update(self.member_repo, consts.MEMBERS,
                                            member, delete_record=True)

            health_mons = status.pop(consts.HEALTHMONITORS, [])
            for health_mon in health_mons:
                self._process_status_update(
                    self.health_mon_repo, consts.HEALTHMONITORS, health_mon,
                    delete_record=True)

            pools = status.pop(consts.POOLS, [])
            for pool in pools:
                self._process_status_update(self.pool_repo, consts.POOLS,
                                            pool, delete_record=True)

            l7rules = status.pop(consts.L7RULES, [])
            for l7rule in l7rules:
                self._process_status_update(self.l7rule_repo, consts.L7RULES,
                                            l7rule, delete_record=True)

            l7policies = status.pop(consts.L7POLICIES, [])
            for l7policy in l7policies:
                self._process_status_update(
                    self.l7policy_repo, consts.L7POLICIES, l7policy,
                    delete_record=True)

            listeners = status.pop(lib_consts.LISTENERS, [])
            for listener in listeners:
                self._process_status_update(
                    self.listener_repo, lib_consts.LISTENERS, listener,
                    delete_record=True)

            lbs = status.pop(consts.LOADBALANCERS, [])
            for lb in lbs:
                self._process_status_update(self.loadbalancer_repo,
                                            consts.LOADBALANCERS, lb)
        except driver_exceptions.UpdateStatusError as e:
            return {lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
                    lib_consts.FAULT_STRING: e.fault_string,
                    lib_consts.STATUS_OBJECT: e.status_object,
                    lib_consts.STATUS_OBJECT_ID: e.status_object_id}
        except Exception as e:
            return {lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
                    lib_consts.FAULT_STRING: str(e)}
        return {lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_OK}

    def update_listener_statistics(self, statistics):
        """Update listener statistics.

        :param statistics: Statistics for listeners:
              id (string): ID for listener.
              active_connections (int): Number of currently active connections.
              bytes_in (int): Total bytes received.
              bytes_out (int): Total bytes sent.
              request_errors (int): Total requests not fulfilled.
              total_connections (int): The total connections handled.
        :type statistics: dict
        :raises: UpdateStatisticsError
        :returns: None
        """
        listener_stats = statistics.get(lib_consts.LISTENERS, [])
        stats_objects = []
        for stat in listener_stats:
            try:
                stats_obj = data_models.ListenerStatistics(
                    listener_id=stat['id'],
                    bytes_in=stat['bytes_in'],
                    bytes_out=stat['bytes_out'],
                    active_connections=stat['active_connections'],
                    total_connections=stat['total_connections'],
                    request_errors=stat['request_errors'],
                    received_time=time.time()
                )
                stats_objects.append(stats_obj)
            except Exception as e:
                return {
                    lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
                    lib_consts.FAULT_STRING: str(e),
                    lib_consts.STATS_OBJECT: lib_consts.LISTENERS}

        # Provider drivers other than the amphora driver do not have
        # an amphora ID, use the listener ID again here to meet the
        # constraint requirement.
        try:
            if stats_objects:
                stats_base.update_stats_via_driver(stats_objects)
        except Exception as e:
            return {
                lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_FAILED,
                lib_consts.FAULT_STRING: str(e),
                lib_consts.STATS_OBJECT: lib_consts.LISTENERS}
        return {lib_consts.STATUS_CODE: lib_consts.DRVR_STATUS_CODE_OK}
